{*********************************************************}
{File:      NCOciBuff.PAS                                 }
{Revision:  0.01.04 / 05.11.1999                          }
{Comment:   NC OCI8 VCL: Oracle8 handles wrapping classes }
{Copyright: (c) 1999, Nica-Com Ltd.                       }
{Author:    Dmitry Arefiev, diman@ncom.ru                 }
{*********************************************************}
{$I NCOciDef.inc}
{*$DEFINE DEBUG}

unit NCOciWrapper;

interface

Uses NCOci, SysUtils, Classes, Windows;

const
    OCI_DEF_INRECDATASIZE = 255;
    OCI_DEF_INITLONGSIZE = 4000;
    OCI_DEF_INITSTRSIZE = 255;
    OCI_DEF_MAXDLP2STRSIZE = 255;
    OCI_DEF_MAXDLP4STRSIZE = 8192;
    OCI_DEF_PIECEBUFFLEN = 4000;
    OCI_DEF_ROWIDLEN = 18;

type
    TOCIHandle = class;
    TOCIEnvironment = class;
    TOCIErrorMessage = class;
    EOCINativeError = class;
    TOCIError = class;
    TOCIService = class;
    TOCISession = class;
    TOCIServer = class;
    TOCIDescriptor = class;
    TOCITransaction = class;
    TOCIStatement = class;
    TOCIVariable = class;
    TOCISelectItem = class;
    TOCILobLocator = class;
    TOCILobLocatorClass = class of TOCILobLocator;
    TOCIIntLocator = class;
    TOCIExtLocator = class;
    TOCIDescribe = class;

    TOCICustomWrappedComponent = class(TOCICustomComponent);
    TOCICustomWrappedDataSet = class(TOCICustomDataSet);

    PInteger = ^Integer;
    PDouble = ^Double;

// --------------------------------------------------------------------------
// HANDLE's & DESCRIPTOR's
// --------------------------------------------------------------------------

    TOCIHandle = class
    private
        FHandle: pOCIHandle;
        FType: ub4;
        FOwner: TOCIHandle;
        FError: TOCIError;
        constructor Create;
        function Init(AEnv: TOCIHandle; AType: ub4): pOCIHandle;
        function GetError: TOCIError;
        procedure Check(const AMsg: String; ACode: sb4);
    public
        destructor Destroy; override;
        procedure SetStringAttr(AAtrType: Integer; const AValue: String);
        function GetStringAttr(AAtrType: Integer): String;
        procedure SetHandleAttr(AAtrType: Integer; AValue: pOCIHandle);
        function GetHandleAttr(AAtrType: Integer): pOCIHandle;
        procedure SetUB1Attr(AAtrType: Integer; AValue: ub1);
        function GetUB1Attr(AAtrType: Integer): ub1;
        procedure SetUB2Attr(AAtrType: Integer; AValue: ub2);
        function GetUB2Attr(AAtrType: Integer): ub2;
        procedure SetUB4Attr(AAtrType: Integer; AValue: ub4);
        function GetUB4Attr(AAtrType: Integer): ub4;
        procedure SetSB1Attr(AAtrType: Integer; AValue: sb1);
        function GetSB1Attr(AAtrType: Integer): sb1;
        procedure SetSB4Attr(AAtrType: Integer; AValue: Sb4);
        function GetSB4Attr(AAtrType: Integer): sb4;
        property Value: pOCIHandle read FHandle;
        property HType: ub4 read FType;
        property Error: TOCIError read GetError;
        property Owner: TOCIHandle read FOwner;
    end;

    TOCIDatabaseMode = (dmThreaded, dmObject, dmEvents, dmShared);
    TOCIDatabaseModes = set of TOCIDatabaseMode;
    TOCIEnvironment = class(TOCIHandle)
    private
        FError: TOCIError;
    public
        constructor Create(AUseThread, AUseObjects, AUseEvents, AUseShare: Boolean);
        property CACHE_ARRAYFLUSH: sb4 index OCI_ATTR_CACHE_ARRAYFLUSH read GetSb4Attr write SetSb4Attr;
        property CACHE_MAX_SIZE: ub4 index OCI_ATTR_CACHE_MAX_SIZE read GetUb4Attr write SetUb4Attr;
        property CACHE_OPT_SIZE: ub4 index OCI_ATTR_CACHE_OPT_SIZE read GetUb4Attr write SetUb4Attr;
        property OBJECT_: sb4 index OCI_ATTR_OBJECT read GetSb4Attr;
        property PINOPTION: ub4 index OCI_ATTR_PINOPTION read GetUb4Attr write SetUb4Attr;
        property ALLOC_DURATION: ub2 index OCI_ATTR_ALLOC_DURATION read GetUb2Attr write SetUb2Attr;
        property PIN_DURATION: ub2 index OCI_ATTR_PIN_DURATION read GetUb2Attr write SetUb2Attr;
        property HEAPALLOC: ub4 index OCI_ATTR_HEAPALLOC read GetUb4Attr;
        property OBJECT_NEWNOTNULL: sb4 index OCI_ATTR_OBJECT_NEWNOTNULL read GetSb4Attr write SetSb4Attr;
        property OBJECT_DETECTCHANGE: sb4 index OCI_ATTR_OBJECT_DETECTCHANGE read GetSb4Attr write SetSb4Attr;
        property SHARED_HEAPALLOC: ub4 index OCI_ATTR_SHARED_HEAPALLOC read GetUb4Attr;
    end;

    TOCIErrorMessage = class
    private
        FErrorCode: sb4;
        FMessage: string;
        FLevel: ub4;
        constructor Create(AOwner: EOCINativeError;
            ALevel: ub4; AErrorCode: sb4; AMessage: PChar);
    public
        property ErrorCode: sb4 read FErrorCode;
        property Message: string read FMessage;
        property Level: ub4 read FLevel;
    end;

    EOCINativeError = class(EOCIErrorBase)
    private
        FErrors: TList;
        FParseErrorOffset: ub4;
        function GetError(Index: Integer): TOCIErrorMessage;
        function GetErrorCount: Integer;
        constructor Create(const AMessage: String; AError: TOCIError);
    public
        destructor Destroy; override;
        property ErrorCount: Integer read GetErrorCount;
        property Errors[Index: Integer]: TOCIErrorMessage read GetError;
        property ParseErrorOffset: ub4 read FParseErrorOffset;
    end;

    TOCIError = class(TOCIHandle)
    public
        constructor Create(AEnv: TOCIEnvironment);
        procedure Check(const AMsg: String; ACode: sword);
        procedure DbgOut(const AMsg: String);
        property DML_ROW_OFFSET: ub4 index OCI_ATTR_DML_ROW_OFFSET read GetUb4Attr;
    end;

    TOCIService = class(TOCIHandle)
    private
        FOnYield: TNotifyEvent;
        procedure DoYield;
        procedure SetNonblockingMode(const Value: Boolean);
        function GetNonblockingMode: Boolean;
    public
        constructor Create(AEnv: TOCIEnvironment);
        procedure Break;
        procedure Reset;
        property NONBLOCKING_MODE: Boolean read GetNonblockingMode write SetNonblockingMode;
        property OnYield: TNotifyEvent read FOnYield write FOnYield;
        property ENV: pOCIHandle index OCI_ATTR_ENV read GetHandleAttr;
        property SERVER: pOCIHandle index OCI_ATTR_SERVER read GetHandleAttr write SetHandleAttr;
        property SESSION: pOCIHandle index OCI_ATTR_SESSION read GetHandleAttr write SetHandleAttr;
        property TRANS: pOCIHandle index OCI_ATTR_TRANS read GetHandleAttr write SetHandleAttr;
        property IN_V8_MODE: ub1 index OCI_ATTR_IN_V8_MODE read GetUb1Attr;
    end;

    TOCIAuthentMode = (amDefault, amSysDba, amSysOper);
    TOCISession = class(TOCIHandle)
    public
        constructor Create(AService: TOCIService);
        destructor Destroy; override;
        procedure Start(const AUser, APassword: String; AAuthentMode: TOCIAuthentMode);
        procedure ChangePassword(const AUser, AOldPassword, ANewPassword: String);
        procedure Stop;
        procedure Select;
        property USERNAME: String index OCI_ATTR_USERNAME read GetStringAttr write SetStringAttr;
        property PASSWORD: String index OCI_ATTR_PASSWORD read GetStringAttr write SetStringAttr;
        property MIGSESSION: ub1 index OCI_ATTR_MIGSESSION read GetUb1Attr write SetUb1Attr;
    end;

    TOCIServer = class(TOCIHandle)
    private
        function GetServerVersion: String;
        procedure SetNonblockingMode(const Value: Boolean);
        function GetNonblockingMode: Boolean;
    public
        constructor Create(AService: TOCIService);
        destructor Destroy; override;
        procedure Attach(const AServerName: String);
        procedure Detach;
        procedure Select;
        property ServerVersion: String read GetServerVersion;
        property NONBLOCKING_MODE: Boolean read GetNonblockingMode write SetNonblockingMode;
        property ENV: pOCIHandle index OCI_ATTR_ENV read GetHandleAttr;
        property EXTERNAL_NAME: String index OCI_ATTR_EXTERNAL_NAME read GetStringAttr write SetStringAttr;
        property INTERNAL_NAME: String index OCI_ATTR_INTERNAL_NAME read GetStringAttr write SetStringAttr;
        property IN_V8_MODE: ub1 index OCI_ATTR_IN_V8_MODE read GetUb1Attr;
        property FOCBK: Pointer index OCI_ATTR_FOCBK read GetHandleAttr write SetHandleAttr;
        property SERVER_GROUP: String index OCI_ATTR_SERVER_GROUP read GetStringAttr write SetStringAttr;
    end;

    TOCITransactionMode = (tmReadWrite, tmSerializable, tmReadOnly);
    TOCITransaction = class(TOCIHandle)
    public
        constructor Create(AService: TOCIService);
        procedure Start(AMode: TOCITransactionMode);
        procedure Commit;
        procedure RollBack;
        procedure Select;
        property TRANS_NAME: String index OCI_ATTR_TRANS_NAME read GetStringAttr write SetStringAttr;
        property XID: Pointer index OCI_ATTR_XID read GetHandleAttr write SetHandleAttr;
    end;

    TOCIPieceLoopData = record
        lastPieceVarHndl: pOCIHandle;
        lastPieceVar: TOCIVariable;
        lastPieceSize: ub4;
        lastInd: sb2;
        lastIter: ub4;
    end;

    TOCIStatement = class(TOCIHandle)
    private
        FEOF: Boolean;
        FLastRowCount: sb4;
        FPieceBuffOwn: Boolean;
        FPieceBuffLen: ub4;
        FPieceBuff: pUb1;
        FPieceOutVars: TList;
        FService: TOCIService;
        procedure _Execute(AService: TOCIService; ARows: sb4;
            ADescrOnly, AExact, ACommit, AScrollable: Boolean);
        procedure Fetch(ANRows: sb4; AOrient: ub2);
        procedure SetPieceBuffOwn(AValue: Boolean);
        procedure SetPieceBuffLen(AValue: ub4);
        procedure AllocPieceBuff;
        procedure FreePieceBuff;
        function Hndl2PieceVar(varHndl: pOCIHandle; varType: ub4; varInOut: ub1): TOCIVariable;
        function GetPiecedFetch: Boolean;
        procedure InitPiecedFetch(var pld: TOCIPieceLoopData);
        procedure DoPiecedFetch(var APieceLoopData: TOCIPieceLoopData; ARes: sb4);
    protected
        procedure AddPieceVar(AVar: TOCIVariable);
        procedure RemovePieceVar(AVar: TOCIVariable);
    public
        constructor Create(AEnv: TOCIEnvironment);
        destructor Destroy; override;
        procedure Prepare(const AStatement: String);
        procedure Describe(AService: TOCIService);
        procedure Execute(AService: TOCIService; ARows: ub4;
            AExact, ACommit, AScrollable: Boolean);
        procedure CancelCursor;
        procedure Next(ANRows: sb4);
        property PieceBuffOwn: Boolean read FPieceBuffOwn write SetPieceBuffOwn;
        property PieceBuffLen: ub4 read FPieceBuffLen write SetPieceBuffLen;
        property EOF: Boolean read FEOF;
        property LastRowCount: sb4 read FLastRowCount;
        property PiecedFetch: Boolean read GetPiecedFetch;
        property NUM_DML_ERRORS: ub4 index OCI_ATTR_NUM_DML_ERRORS read GetUb4Attr;
        property ENV: pOCIHandle index OCI_ATTR_ENV read GetHandleAttr;
        property STMT_TYPE: ub2 index OCI_ATTR_STMT_TYPE read GetUB2Attr;
        property ROW_COUNT: ub4 index OCI_ATTR_ROW_COUNT read GetUB4Attr;
        property SQLFNCODE: ub2 index OCI_ATTR_SQLFNCODE read GetUB2Attr;
        property ROWID: pOCIRowid index OCI_ATTR_ROWID read GetHandleAttr;
        property PARAM_COUNT: ub4 index OCI_ATTR_PARAM_COUNT read GetUB4Attr;
        property PREFETCH_ROWS: ub4 index OCI_ATTR_PREFETCH_ROWS read GetUB4Attr write SetUB4Attr;
        property PREFETCH_MEMORY: ub4 index OCI_ATTR_PREFETCH_MEMORY read GetUB4Attr write SetUB4Attr;
        property PARSE_ERROR_OFFSET: ub4 index OCI_ATTR_PARSE_ERROR_OFFSET read GetUB4Attr;
    end;

    TOCIVarDataType = (otUnknown, otSmallInt, otInteger, otFloat, otBCD, otString,
                       otChar, otRaw, otDateTime, otLong, otLongRaw, otCursor,
                       otCLOB, otBLOB, otCFile, otBFile, otROWID);
    TOCIVarDataTypes = set of TOCIVarDataType;
    TOCIVarType = (odUnknown, odIn, odOut, odInOut, odRet, odDefine);

    TBindData = record
        FInd: sb2;
        FALen: ub2;
        FRCode: ub2;
        FBuffer: record end;
    end;
    PBindData = ^TBindData;

    TOraDate = record
        century, year, month, day,
        hour, minute, second: ub1;
    end;
    POraDate = ^TOraDate;

    TOraLong = record
        FSize: sb4;
        FData: pUb1;
    end;
    POraLong = ^TOraLong;

    TOCIBuffStatus = (bsInt, bsExtUnknown, bsExtInit);

    TOCIVariable = class(TOCIHandle)
    private
        FPlaceHolder: String;
        FPosition: sb4;
        FValue_p: PBindData;  // values buffer
        FValue_sz: sb4;       // value max length
        FValue_ty: ub2;       // value type
        FMaxArr_len: sb4;     // maximum number of elements
        FCurEle: sb4;         // actual number of elements
        FDataType: TOCIVarDataType;
        FVarType: TOCIVarType;
        FBuffStatus: TOCIBuffStatus;
        FInrecDataSize: sb4;
        FIsPLSQLTable: Boolean;
        FLongData: Boolean;
        procedure SetArrayLen(const Value: sb4);
        procedure SetDataSize(const Value: sb4);
        procedure SetDataType(const Value: TOCIVarDataType);
        procedure SetVarType(const Value: TOCIVarType);
        function GetBuffer(AIndex: sb4): pUb1;
        procedure FreeBuffer;
        procedure CheckBufferAccess(AIndex: sb4);
        function GetBindRecLen: sb4;
        function GetPBindData(AIndex: sb4): PBindData;
        procedure CheckCanWrite;
        procedure SetExternBuffer(AValue: pUb1);
        function GetExternBuffer: pUb1;
        procedure FreeLong(pLong: POraLong);
        procedure InitLong(pLong: POraLong);
        procedure GetLong(pLong: POraLong; var AData: pUb1; var ASize: sb4);
        procedure SetLong(pLong: POraLong; AData: pUb1; ASize: sb4);
        procedure AppendLong(pLong: POraLong; AData: pUb1; ASize: sb4);
        procedure InitLongData;
        procedure SetInrecDataSize(const Value: sb4);
        procedure CheckOwner;
    protected
        property Buffer[AIndex: sb4]: pUb1 read GetBuffer;
    public
        constructor Create(AStmt: TOCIStatement);
        destructor Destroy; override;
        procedure Bind;
        // for TOciParam (parameters)
        procedure BindTo(AStmt: TOCIStatement);
        procedure BindOff;
        // for TOCIDeadCursor (external buffer management)
        procedure ClearBuffer(AIndex: ub4 = MAXINT; ANoHandleFree: Boolean = False);
        procedure InitBuffer(AIndex: ub4 = MAXINT; ANoHandleFree: Boolean = False);
        // Value related
        function GetDataPtr(AIndex: ub4; var ABuff: pUb1; var ASize: sb4): Boolean;
        function GetData(AIndex: ub4; ABuff: pUb1; var ASize: sb4): Boolean;
        procedure SetData(AIndex: ub4; ABuff: pUb1; ASize: sb4);
        procedure AppendData(AIndex: ub4; ABuff: pUb1; ASize: sb4);
        procedure SetIsNull(AIndex: ub4; AIsNull: Boolean); // for LOB fields
        // -----------
        property Name: String read FPlaceHolder write FPlaceHolder;
        property Position: sb4 read FPosition write FPosition;
        property IsPLSQLTable: Boolean read FIsPLSQLTable write FIsPLSQLTable;
        property DataType: TOCIVarDataType read FDataType write SetDataType;
        property DataSize: sb4 read FValue_sz write SetDataSize;
        property VarType: TOCIVarType read FVarType write SetVarType;
        property ArrayLen: sb4 read FMaxArr_len write SetArrayLen;
        property ArrayCount: sb4 read FCurEle;
        // -----------
        property ExternBuffer: pUb1 read GetExternBuffer write SetExternBuffer;
        property BindRecLen: sb4 read GetBindRecLen;
        property InrecDataSize: sb4 read FInrecDataSize write SetInrecDataSize;
        property LongData: Boolean read FLongData;
        // -----------
        property CHAR_COUNT: ub4 index OCI_ATTR_CHAR_COUNT read GetUb4Attr write SetUb4Attr;
        property CHARSET_ID: ub2 index OCI_ATTR_CHARSET_ID read GetUb2Attr write SetUb2Attr;
        property CHARSET_FORM: ub1 index OCI_ATTR_CHARSET_FORM read GetUb1Attr write SetUb1Attr;
        property PDPRC: ub2 index OCI_ATTR_PDPRC read GetUb2Attr write SetUb2Attr;
        property PDSCL: ub2 index OCI_ATTR_PDSCL read GetUb2Attr write SetUb2Attr;
        property ROWS_RETURNED: ub4 index OCI_ATTR_ROWS_RETURNED read GetUB4Attr;
        property MAXDATA_SIZE: sb4 index OCI_ATTR_MAXDATA_SIZE read GetSb4Attr write SetSb4Attr;
    end;

    TOCIDescriptor = class(TOCIHandle)
    private
        function Init(AEnv: TOCIHandle; AType: ub4): pOCIDescriptor;
    public
        destructor Destroy; override;
    end;

    TOCISelectItem = class(TOCIDescriptor)
    private
        FPosition: sb4;
        FEnableInteger: Boolean;
        FEnableBCD: Boolean;
        FEnableLongString: Boolean;
        FEnableFixedString: Boolean;
        function GetDataType: TOCIVarDataType;
        function GetDataSize: ub4;
        function GetVarType: TOCIVarType;
    public
        constructor Create(AStmt: TOCIStatement);
        destructor Destroy; override;
        procedure Bind;
        property Position: sb4 read FPosition write FPosition;
        property DataType: TOCIVarDataType read GetDataType;
        property DataSize: ub4 read GetDataSize;
        property VarType: TOCIVarType read GetVarType;
        property EnableInteger: Boolean read FEnableInteger write FEnableInteger;
        property EnableBCD: Boolean read FEnableBCD write FEnableBCD;
        property EnableLongString: Boolean read FEnableLongString write FEnableLongString;
        property EnableFixedString: Boolean read FEnableFixedString write FEnableFixedString;
        property DATA_SIZE: ub4 index OCI_ATTR_DATA_SIZE read GetUb4Attr;
        property DATA_TYPE: ub4 index OCI_ATTR_DATA_TYPE read GetUb4Attr;
        property DISP_SIZE: ub4 index OCI_ATTR_DISP_SIZE read GetUb4Attr;
        property NAME: String index OCI_ATTR_NAME read GetStringAttr;
        property PRECISION: ub4 index OCI_ATTR_PRECISION read GetUb4Attr;
        property SCALE: ub4 index OCI_ATTR_SCALE read GetUb4Attr;
        property IS_NULL: ub4 index OCI_ATTR_IS_NULL read GetUb4Attr;
        property TYPE_NAME: String index OCI_ATTR_TYPE_NAME read GetStringAttr;
        property SUB_NAME: String index OCI_ATTR_SUB_NAME read GetStringAttr;
        property SCHEMA_NAME: String index OCI_ATTR_SCHEMA_NAME read GetStringAttr;
        property IOMODE: ub1 index OCI_ATTR_IOMODE read GetUB1Attr;
    end;

    TOCILobLocator = class(TOCIDescriptor)
    private
        FOwned: Boolean;
        function GetLength: sb4;
        function GetIsOpen: LongBool;
        function GetIsInit: LongBool;
    public
        constructor Create(AService: TOCIService; AType: ub4);
        constructor CreateUseHandle(AService: TOCIService; AHandle: pOCIHandle); virtual;
        destructor Destroy; override;
        procedure Assign(AFrom: TOCILobLocator);
        procedure Close;
        procedure Open(AReadOnly: Boolean);
        procedure Read(ABuff: pUb1; ABuffLen: ub4; var amount: ub4; offset: ub4);
        property Length: sb4 read GetLength;
        property IsOpen: LongBool read GetIsOpen;
        property IsInit: LongBool read GetIsInit;
    end;

    TOCIIntLocator = class(TOCILobLocator)
    private
        FBuffering: Boolean;
        procedure SetBuffering(const Value: Boolean);
    public
        constructor CreateUseHandle(AService: TOCIService; AHandle: pOCIHandle); override;
        procedure Append(AFrom: TOCIIntLocator);
        procedure Copy(AFrom: TOCIIntLocator; amount, dst_offset, src_offset: ub4);
        procedure Erase(var amount: ub4; offset: ub4);
        procedure FlushBuffer;
        procedure LoadFromFile(AFrom: TOCIExtLocator; amount, dst_offset, src_offset: ub4);
        procedure Trim(ANewLen: ub4);
        procedure Write(ABuff: pUb1; ABuffLen: ub4; var amount: ub4; offset: ub4);
        property Buffering: Boolean read FBuffering write SetBuffering;
    end;

    TOCIExtLocator = class(TOCILobLocator)
    private
        function GetFileExists: Boolean;
        function GetDirectory: String;
        function GetFileName: String;
        procedure SetDirectory(const Value: String);
        procedure SetFileName(const Value: String);
        procedure GetFileDir(var ADir, AFileName: String);
        procedure SetFileDir(const ADir, AFileName: String);
    public
        constructor CreateUseHandle(AService: TOCIService; AHandle: pOCIHandle); override;
        property FileExists: Boolean read GetFileExists;
        property FileName: String read GetFileName write SetFileName;
        property Directory: String read GetDirectory write SetDirectory;
    end;

    TOCIDescribe = class(TOCIHandle)
    private
        FCurr: TOCIHandle;
        FStack: TList;
        function GetPtr(AIndex: Integer): Pointer;
        function GetSB1(AIndex: Integer): sb1;
        function GetText(AIndex: Integer): String;
        function GetUB1(AIndex: Integer): ub1;
        function GetUB2(AIndex: Integer): ub2;
        function GetUB4(AIndex: Integer): ub4;
        function GetSelectItem(AIndex: Integer): TOCISelectItem;
    public
        constructor Create(ASvc: TOCIService);
        destructor Destroy; override;
        procedure DescribeName(const AName: String);
        function OpenList(AAtrType: Integer): Integer;
        procedure CloseList;
        procedure GoToItem(AIndex: Integer);
        property SB1[AIndex: Integer]: sb1 read GetSB1;
        property UB1[AIndex: Integer]: ub1 read GetUB1;
        property UB2[AIndex: Integer]: ub2 read GetUB2;
        property UB4[AIndex: Integer]: ub4 read GetUB4;
        property TEXT[AIndex: Integer]: String read GetText;
        property PTR[AIndex: Integer]: Pointer read GetPtr;
        property SelectItem[AIndex: Integer]: TOCISelectItem read GetSelectItem;
    end;

implementation

Uses NCOciMsg;

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------

const
    hndlNames: array[1..64] of String =
   ('OCI_HTYPE_ENV',
    'OCI_HTYPE_ERROR',
    'OCI_HTYPE_SVCCTX',
    'OCI_HTYPE_STMT',
    'OCI_HTYPE_BIND',
    'OCI_HTYPE_DEFINE',
    'OCI_HTYPE_DESCRIBE',
    'OCI_HTYPE_SERVER',
    'OCI_HTYPE_SESSION',
    'OCI_HTYPE_TRANS',
    'OCI_HTYPE_COMPLEXOBJECT',
    'OCI_HTYPE_SECURITY',
    'OCI_HTYPE_SUBSCRIPTION',
    'OCI_HTYPE_DIRPATH_CTX',
    'OCI_HTYPE_DIRPATH_COLUMN_ARRAY',
    'OCI_HTYPE_DIRPATH_STREAM',
    'OCI_HTYPE_PROC',
    '','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','','',
    'OCI_DTYPE_LOB',
    'OCI_DTYPE_SNAP',
    'OCI_DTYPE_RSET',
    'OCI_DTYPE_PARAM',
    'OCI_DTYPE_ROWID',
    'OCI_DTYPE_COMPLEXOBJECTCOMP',
    'OCI_DTYPE_FILE',
    'OCI_DTYPE_AQENQ_OPTIONS',
    'OCI_DTYPE_AQDEQ_OPTIONS',
    'OCI_DTYPE_AQMSG_PROPERTIES',
    'OCI_DTYPE_AQAGENT',
    'OCI_DTYPE_LOCATOR',
    'OCI_DTYPE_DATETIME',
    'OCI_DTYPE_INTERVAL',
    'OCI_DTYPE_AQNFY_DESCRIPTOR'
   );

constructor TOCIHandle.Create;
begin
    inherited Create;
    FHandle := nil;
    FType := 0;
    FOwner := nil;
end;

destructor TOCIHandle.Destroy;
begin
    if FHandle <> nil then
        Check('OCIHandleFree(' + hndlNames[FType] + ')',
            NCOci.OCIHandleFree(FHandle, FType));
    inherited Destroy;
end;

function TOCIHandle.Init(AEnv: TOCIHandle; AType: ub4): pOCIHandle;
begin
    Check('OCIHandleAlloc(' + hndlNames[AType] + ')',
        NCOci.OCIHandleAlloc(AEnv.FHandle, FHandle, AType, 0, nil));
    FType := AType;
    Result := @FHandle;
end;

procedure TOCIHandle.SetStringAttr(AAtrType: Integer; const AValue: String);
begin
    Check('OCIAttrSet Str', NCOci.OCIAttrSet(FHandle, FType, PChar(AValue),
        Length(AValue), AAtrType, Error.FHandle));
end;

function TOCIHandle.GetStringAttr(AAtrType: Integer): String;
var
    l: Integer;
    p: PChar;
begin
    Check('OCIAttrGet Str', NCOci.OCIAttrGet(FHandle, FType, @p, @l,
        AAtrType, Error.FHandle));
    SetString(Result, p, l);
end;

procedure TOCIHandle.SetHandleAttr(AAtrType: Integer; AValue: pOCIHandle);
begin
    Check('OCIAttrSet Hndl', NCOci.OCIAttrSet(FHandle, FType, AValue, 0,
        AAtrType, Error.FHandle));
end;

function TOCIHandle.GetHandleAttr(AAtrType: Integer): pOCIHandle;
begin
    Result := nil;
    Check('OCIAttrGet Hndl', NCOci.OCIAttrGet(FHandle, FType, @Result, nil,
        AAtrType, Error.FHandle));
end;

procedure TOCIHandle.SetUB1Attr(AAtrType: Integer; AValue: ub1);
begin
    Check('OCIAttrSet ub1', NCOci.OCIAttrSet(FHandle, FType, @AValue, 0,
        AAtrType, Error.FHandle));
end;

function TOCIHandle.GetUB1Attr(AAtrType: Integer): ub1;
begin
    Result := 0;
    Check('OCIAttrGet ub1', NCOci.OCIAttrGet(FHandle, FType, @Result, nil,
        AAtrType, Error.FHandle));
end;

procedure TOCIHandle.SetUB2Attr(AAtrType: Integer; AValue: ub2);
begin
    Check('OCIAttrSet ub2', NCOci.OCIAttrSet(FHandle, FType, @AValue, 0,
        AAtrType, Error.FHandle));
end;

function TOCIHandle.GetUB2Attr(AAtrType: Integer): ub2;
begin
    Result := 0;
    Check('OCIAttrGet ub2', NCOci.OCIAttrGet(FHandle, FType, @Result, nil,
        AAtrType, Error.FHandle));
end;

procedure TOCIHandle.SetUB4Attr(AAtrType: Integer; AValue: ub4);
begin
    Check('OCIAttrSet ub4', NCOci.OCIAttrSet(FHandle, FType, @AValue, 0,
        AAtrType, Error.FHandle));
end;

function TOCIHandle.GetUB4Attr(AAtrType: Integer): ub4;
begin
    Result := 0;
    Check('OCIAttrGet ub4', NCOci.OCIAttrGet(FHandle, FType, @Result, nil,
        AAtrType, Error.FHandle));
end;

procedure TOCIHandle.SetSB1Attr(AAtrType: Integer; AValue: Sb1);
begin
    Check('OCIAttrSet sb1', NCOci.OCIAttrSet(FHandle, FType, @AValue, 0,
        AAtrType, Error.FHandle));
end;

function TOCIHandle.GetSB1Attr(AAtrType: Integer): sb1;
begin
    Result := 0;
    Check('OCIAttrGet sb1', NCOci.OCIAttrGet(FHandle, FType, @Result, nil,
        AAtrType, Error.FHandle));
end;

procedure TOCIHandle.SetSB4Attr(AAtrType: Integer; AValue: Sb4);
begin
    Check('OCIAttrSet sb4', NCOci.OCIAttrSet(FHandle, FType, @AValue, 0,
        AAtrType, Error.FHandle));
end;

function TOCIHandle.GetSB4Attr(AAtrType: Integer): sb4;
begin
    Result := 0;
    Check('OCIAttrGet sb4', NCOci.OCIAttrGet(FHandle, FType, @Result, nil,
        AAtrType, Error.FHandle));
end;

function TOCIHandle.GetError: TOCIError;
var
    h: TOCIHandle;
begin
    Result := FError;
    if Result = nil then begin
        h := Self;
        while (h <> nil) and not (h is TOCIEnvironment) do
            h := h.Owner;
        FError := TOCIEnvironment(h).FError;
        Result := FError;
    end;
end;

procedure TOCIHandle.Check(const AMsg: String; ACode: sb4);
begin
{$IFDEF DEBUG}
    Error.DbgOut(AMsg);
{$ENDIF}    
    if ACode <> OCI_SUCCESS then
        Error.Check(AMsg, ACode);
end;

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------

constructor TOCIEnvironment.Create(AUseThread, AUseObjects, AUseEvents, AUseShare: Boolean);
var
    mode: ub4;
    err1, err2: sword;
begin
    inherited Create;
    FType := OCI_HTYPE_ENV;
    mode := OCI_DEFAULT;
    if AUseThread then
        mode := mode or OCI_THREADED;
    if AUseObjects then
        mode := mode or OCI_OBJECT;
    if (FOCIVersion >= cvOracle8i) and AUseEvents then
        mode := mode or OCI_EVENTS;
    if (FOCIVersion >= cvOracle8i) and AUseShare then
        mode := mode or OCI_SHARED;
    InitOCI;
    if FOCIVersion >= cvOracle8i then begin
        err1 := NCOci.OCIEnvCreate(FHandle, mode, nil, nil, nil, nil, 0, nil);
        err2 := OCI_SUCCESS;
    end
    else begin
        err1 := NCOci.OCIInitialize(mode, nil, nil, nil, nil);
        err2 := NCOci.OCIEnvInit(FHandle, OCI_DEFAULT, 0, nil);
    end;
    FError := TOCIError.Create(Self);
    if FOCIVersion >= cvOracle8i then
        Check('OCIEnvCreate', err1)
    else begin
        Check('OCIInitialize', err1);
        Check('OCIEnvInit', err2);
    end;
end;

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------

constructor TOCIErrorMessage.Create(AOwner: EOCINativeError;
    ALevel: ub4; AErrorCode: sb4; AMessage: PChar);
begin
    inherited Create;
    AOwner.FErrors.Add(Self);
    FLevel := ALevel;
    FErrorCode := AErrorCode;
    FMessage := AMessage;
end;

constructor EOCINativeError.Create(const AMessage: String; AError: TOCIError);
var
    lvl: ub4;
    errCode: sb4;
    errBuf: array[0..511] of Char;
begin
    inherited Create(AMessage);
    FErrors := TList.Create;
    lvl := 1;
{$IFDEF DEBUG}
    AError.DbgOut('ERROR> ' + AMessage);
{$ENDIF}    
    while NCOci.OCIErrorGet(AError.FHandle, lvl, nil, errCode, errBuf, sizeof(errBuf),
                          OCI_HTYPE_ERROR) <> OCI_NO_DATA do begin
        TOCIErrorMessage.Create(Self, lvl, errCode, errBuf);
{$IFDEF DEBUG}
        AError.DbgOut('ERROR> ' + errBuf);
{$ENDIF}
        if lvl = 1 then
            Message := Message + #13#10 + errBuf;
        Inc(lvl);
    end;
end;

destructor EOCINativeError.Destroy;
var
    i: Integer;
begin
    for i := 0 to FErrors.Count - 1 do
        TOCIErrorMessage(FErrors[i]).Free;
    FErrors.Free;
    inherited Destroy;
end;

function EOCINativeError.GetError(Index: Integer): TOCIErrorMessage;
begin
    Result := TOCIErrorMessage(FErrors[Index]);
end;

function EOCINativeError.GetErrorCount: Integer;
begin
    Result := FErrors.Count;
end;

constructor TOCIError.Create(AEnv: TOCIEnvironment);
begin
    inherited Create;
    FOwner := AEnv;
    Init(AEnv, OCI_HTYPE_ERROR);
end;

procedure TOCIError.Check(const AMsg: String; ACode: sword);
begin
    case ACode of
    OCI_SUCCESS: ;
    OCI_SUCCESS_WITH_INFO:
        OCIDBError('OCI_SUCCESS_WITH_INFO');
    OCI_NEED_DATA:
        OCIDBError('OCI_NEED_DATA');
    OCI_NO_DATA:
        OCIDBError('OCI_NO_DATA');
    OCI_ERROR:
         raise EOCINativeError.Create('OCI_ERROR', self);
    OCI_INVALID_HANDLE:
        OCIDBError('OCI_INVALID_HANDLE');
    OCI_STILL_EXECUTING:
        OCIDBError('OCI_STILL_EXECUTING');
    OCI_CONTINUE:
        OCIDBError('OCI_CONTINUE');
    else ;
    end;
end;

procedure TOCIError.DbgOut(const AMsg: String);
begin
    OutputDebugString(PChar(AMsg));
end;

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------

constructor TOCIService.Create(AEnv: TOCIEnvironment);
begin
    inherited Create;
    FOwner := AEnv;
    Init(AEnv, OCI_HTYPE_SVCCTX);
end;

procedure TOCIService.Break;
begin
    Check('OCIBreak', NCOci.OCIBreak(FHandle, Error.FHandle));
end;

procedure TOCIService.Reset;
begin
    Check('OCIReset', NCOci.OCIReset(FHandle, Error.FHandle));
end;

procedure TOCIService.DoYield;
begin
    if Assigned(FOnYield) then
        FOnYield(Self);
end;

procedure TOCIService.SetNonblockingMode(const Value: Boolean);
begin
    while NONBLOCKING_MODE <> Value do
        SetUb1Attr(OCI_ATTR_NONBLOCKING_MODE, ub1(Value));
end;

function TOCIService.GetNonblockingMode: Boolean;
begin
    Result := Boolean(GetUb1Attr(OCI_ATTR_NONBLOCKING_MODE));
end;

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------

constructor TOCIServer.Create(AService: TOCIService);
begin
    inherited Create;
    FOwner := AService;
    Init(AService.FOwner, OCI_HTYPE_SERVER);
end;

destructor TOCIServer.Destroy;
begin
    try
        Detach;
    except
    end;        
    inherited Destroy;
end;

procedure TOCIServer.Attach(const AServerName: String);
begin
    Check('OCIServerAttach(' + AServerName + ')',
        NCOci.OCIServerAttach(FHandle, Error.FHandle,
        PChar(AServerName), Length(AServerName), OCI_DEFAULT));
    Select;
end;

procedure TOCIServer.Detach;
begin
    if FHandle <> nil then
        Check('OCIServerDetach', NCOci.OCIServerDetach(FHandle, Error.FHandle, OCI_DEFAULT));
end;

procedure TOCIServer.Select;
begin
    TOCIService(FOwner).SERVER := FHandle;
end;

function TOCIServer.GetServerVersion: String;
const
    OCI_MAX_ATTR_LEN = 1024;
var
    buff: array[0 .. OCI_MAX_ATTR_LEN-1] of char;
begin
    OCIServerVersion(FHandle, Error.FHandle, @buff, OCI_MAX_ATTR_LEN, OCI_HTYPE_SERVER);
    Result := buff;
end;

procedure TOCIServer.SetNonblockingMode(const Value: Boolean);
begin
    while NONBLOCKING_MODE <> Value do
        SetUb1Attr(OCI_ATTR_NONBLOCKING_MODE, ub1(Value));
end;

function TOCIServer.GetNonblockingMode: Boolean;
begin
    Result := Boolean(GetUb1Attr(OCI_ATTR_NONBLOCKING_MODE));
end;

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------

constructor TOCISession.Create(AService: TOCIService);
begin
    inherited Create;
    FOwner := AService;
    Init(AService.FOwner, OCI_HTYPE_SESSION);
end;

destructor TOCISession.Destroy;
begin
    try
        Stop;
    except
    end;
    inherited Destroy;
end;

procedure TOCISession.Start(const AUser, APassword: String; AAuthentMode: TOCIAuthentMode);
var
    credt: ub4;
    mode: ub4;
begin
    if (AUser = '') and (APassword = '') then
        credt := OCI_CRED_EXT
    else begin
        USERNAME := AUser;
        PASSWORD := APassword;
        credt := OCI_CRED_RDBMS;
    end;
    mode := OCI_DEFAULT;
    if AAuthentMode = amSysDba then
        mode := OCI_SYSDBA
    else if AAuthentMode = amSysOper then
        mode := OCI_SYSOPER;
    Check('OCISessionBegin(' + AUser + ')', NCOci.OCISessionBegin(FOwner.FHandle,
        Error.FHandle, FHandle, credt, mode));
    Select;
end;

procedure TOCISession.ChangePassword(const AUser, AOldPassword, ANewPassword: String);
begin
    Check('OCIPasswordChange', NCOci.OCIPasswordChange(FOwner.FHandle,
        Error.FHandle, PChar(AUser), Length(AUser), PChar(AOldPassword),
        Length(AOldPassword), PChar(ANewPassword), Length(ANewPassword),
        OCI_DEFAULT));
end;

procedure TOCISession.Stop;
begin
    if FHandle <> nil then
        Check('OCISessionEnd', NCOci.OCISessionEnd(FOwner.FHandle, Error.FHandle,
            FHandle, OCI_DEFAULT));
end;

procedure TOCISession.Select;
begin
    TOCIService(FOwner).SESSION := FHandle;
end;

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------

constructor TOCITransaction.Create(AService: TOCIService);
begin
    inherited Create;
    FOwner := AService;
    Init(AService.FOwner, OCI_HTYPE_TRANS);
    Select;
end;

procedure TOCITransaction.Start(AMode: TOCITransactionMode);
const
    nc2OCItm: array[TOCITransactionMode] of ub4 = (OCI_TRANS_READWRITE,
        OCI_TRANS_SERIALIZABLE, OCI_TRANS_READONLY);
begin
    Check('OCITransStart', NCOci.OCITransStart(FOwner.FHandle, Error.FHandle,
        0, nc2OCItm[AMode]));
end;

procedure TOCITransaction.Commit;
begin
    Check('OCITransCommit', NCOci.OCITransCommit(FOwner.FHandle, Error.FHandle,
        OCI_DEFAULT));
end;

procedure TOCITransaction.RollBack;
begin
    Check('OCITransRollback', NCOci.OCITransRollback(FOwner.FHandle, Error.FHandle,
        OCI_DEFAULT));
end;

procedure TOCITransaction.Select;
begin
    TOCIService(FOwner).TRANS := FHandle;
end;

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------

constructor TOCIStatement.Create(AEnv: TOCIEnvironment);
begin
    inherited Create;
    FOwner := AEnv;
    Init(AEnv, OCI_HTYPE_STMT);
    FPieceBuffLen := 4000;
    FPieceOutVars := TList.Create;
end;

destructor TOCIStatement.Destroy;
begin
    FPieceOutVars.Free;
    FreePieceBuff;
    inherited Destroy;
end;

// -----------------------------------------
// Piece wise operation support

var
    FGlobalPieceBuffSize: ub4;
    FGlobalPieceBuff: pUb1;
    FGlobalPieceBuffUseCnt: ub4;

function GetGlobalPieceBuff(ASize: ub4): pUb1;
begin
    if FGlobalPieceBuffUseCnt = 0 then begin
        GetMem(FGlobalPieceBuff, ASize);
        FGlobalPieceBuffSize := ASize;
    end
    else if FGlobalPieceBuffSize < ASize then begin
        ReallocMem(FGlobalPieceBuff, ASize);
        FGlobalPieceBuffSize := ASize;
    end;
    Inc(FGlobalPieceBuffUseCnt);
    Result := FGlobalPieceBuff;
end;

procedure FreeGlobalPieceBuff;
begin
    Dec(FGlobalPieceBuffUseCnt);
    if FGlobalPieceBuffUseCnt = 0 then begin
        FreeMem(FGlobalPieceBuff, FGlobalPieceBuffSize);
        FGlobalPieceBuff := nil;
        FGlobalPieceBuffSize := 0;
    end;
end;

procedure TOCIStatement.AllocPieceBuff;
begin
    if FPieceBuff = nil then
        if FPieceBuffOwn then
            GetMem(FPieceBuff, FPieceBuffLen)
        else
            FPieceBuff := GetGlobalPieceBuff(FPieceBuffLen);
end;

procedure TOCIStatement.FreePieceBuff;
begin
    if FPieceBuff <> nil then
    try
        if FPieceBuffOwn then
            FreeMem(FPieceBuff, FPieceBuffLen)
        else
            FreeGlobalPieceBuff;
    except
        FPieceBuff := nil;
        raise;
    end;
end;

procedure TOCIStatement.SetPieceBuffOwn(AValue: Boolean);
begin
    if FPieceBuffOwn <> AValue then begin
        FreePieceBuff;
        FPieceBuffOwn := AValue;
    end;
end;

procedure TOCIStatement.SetPieceBuffLen(AValue: ub4);
begin
    if FPieceBuffLen <> AValue then begin
        FreePieceBuff;
        FPieceBuffLen := AValue;
    end;
end;

function TOCIStatement.Hndl2PieceVar(varHndl: pOCIHandle; varType: ub4; varInOut: ub1): TOCIVariable;
var
    i: Integer;
begin
    i := 0;
    while (i < FPieceOutVars.Count) and
          (TOCIVariable(FPieceOutVars[i]).FHandle <> varHndl) do
        Inc(i);
    if i = FPieceOutVars.Count then
        OCIDBError(msgOCIUndefPWVar);
    Result := TOCIVariable(FPieceOutVars[i]);
    if not Result.LongData or (varType <> OCI_HTYPE_DEFINE) or
       (varInOut <> OCI_PARAM_OUT) then
        OCIDBError(msgOCIBadTypePWVar);
end;

procedure TOCIStatement.AddPieceVar(AVar: TOCIVariable);
begin
    if FPieceOutVars.IndexOf(AVar) = -1 then
        FPieceOutVars.Add(AVar);
end;

procedure TOCIStatement.RemovePieceVar(AVar: TOCIVariable);
begin
    FPieceOutVars.Remove(AVar);
end;

function TOCIStatement.GetPiecedFetch: Boolean;
begin
    Result := FPieceOutVars.Count > 0;
end;

procedure TOCIStatement.InitPiecedFetch(var pld: TOCIPieceLoopData);
begin
    if PiecedFetch then
        FillChar(pld, SizeOf(TOCIPieceLoopData), 0);
end;

procedure TOCIStatement.DoPiecedFetch(var APieceLoopData: TOCIPieceLoopData; ARes: sb4);
var
    varHndl: pOCIHandle;
    varType: ub4;
    varInOut: ub1;
    idx: ub4;
    piece: ub1;
    rcode: ub2;
begin
    if PiecedFetch then
    with APieceLoopData do begin
        if lastPieceVar <> nil then
            if lastInd = -1 then
                lastPieceVar.SetData(lastIter, nil, 0)
            else
                lastPieceVar.AppendData(lastIter, FPieceBuff, lastPieceSize);
        if ARes = OCI_NEED_DATA then begin
            Check('OCIStmtGetPieceInfo', NCOci.OCIStmtGetPieceInfo(FHandle,
                Error.FHandle, varHndl, varType, varInOut, lastIter, idx, piece));
            if varHndl <> lastPieceVarHndl then begin
                lastPieceVar := Hndl2PieceVar(varHndl, varType, varInOut);
                lastPieceVarHndl := varHndl;
            end;
            AllocPieceBuff;
            lastPieceSize := FPieceBuffLen;
            Check('OCIStmtSetPieceInfo', NCOci.OCIStmtSetPieceInfo(lastPieceVar.FHandle,
                OCI_HTYPE_DEFINE, Error.FHandle, FPieceBuff, lastPieceSize,
                piece, @lastInd, rcode));
        end
        else
            lastPieceVar := nil;
    end;
end;

// -----------------------------------------
// Actual operation's

procedure TOCIStatement.Prepare(const AStatement: String);
begin
    Check('OCIStmtPrepare(' + AStatement + ')', NCOci.OCIStmtPrepare(FHandle,
        Error.FHandle, PChar(AStatement), Length(AStatement), OCI_NTV_SYNTAX,
        OCI_DEFAULT));
end;

procedure TOCIStatement._Execute(AService: TOCIService; ARows: sb4;
    ADescrOnly, AExact, ACommit, AScrollable: Boolean);
var
    mode: ub4;
    res: sb4;
    pieceLoopData: TOCIPieceLoopData;
begin
    mode := OCI_DEFAULT;
    FService := AService;
    if ADescrOnly then
        mode := mode or OCI_DESCRIBE_ONLY;
    if AExact then
        mode := mode or OCI_EXACT_FETCH;
    if ACommit then
        mode := mode or OCI_COMMIT_ON_SUCCESS;
    if Stmt_Type = OCI_STMT_SELECT then begin
        if AScrollable then
            mode := mode or OCI_SCROLLABLE_CURSOR;
    end
    else begin
        if ARows = 0 then
            ARows := 1;
    end;
    FLastRowCount := 0;
    InitPiecedFetch(pieceLoopData);
    try
        repeat
            res := NCOci.OCIStmtExecute(FService.FHandle, FHandle, Error.FHandle,
                ARows, 0, nil, nil, mode);
            case res of
                OCI_STILL_EXECUTING:
                    FService.DoYield;
                OCI_NEED_DATA, OCI_NO_DATA, OCI_SUCCESS:
                    begin
                        if (res = OCI_NO_DATA) and AExact then
                            Check('OCIStmtExecute', res)
                        else
                            Check('OCIStmtExecute', OCI_SUCCESS);
                    end;
                else
                    Check('OCIStmtExecute', res);
            end;
            DoPiecedFetch(pieceLoopData, res);
        until (res <> OCI_NEED_DATA) and (res <> OCI_STILL_EXECUTING);
    except
        on E: EOCINativeError do begin
            if FOCIVersion >= cvOracle8i then
                E.FParseErrorOffset := PARSE_ERROR_OFFSET;
            raise;
        end;
    end;
    FLastRowCount := ROW_COUNT;
    FEOF := FLastRowCount <> ARows;
end;

procedure TOCIStatement.Describe(AService: TOCIService);
begin
    _Execute(AService, 0, True, False, False, False);
end;

procedure TOCIStatement.Execute(AService: TOCIService; ARows: ub4;
    AExact, ACommit, AScrollable: Boolean);
begin
    _Execute(AService, ARows, False, AExact, ACommit, AScrollable);
end;

procedure TOCIStatement.Fetch(ANRows: sb4; AOrient: ub2);
var
    res: sb4;
    prevRowCount: ub4;
    pieceLoopData: TOCIPieceLoopData;
begin
    prevRowCount := ROW_COUNT;
    InitPiecedFetch(pieceLoopData);
    repeat
        res := NCOci.OCIStmtFetch(FHandle, Error.FHandle, ANRows, AOrient, OCI_DEFAULT);
        case res of
            OCI_STILL_EXECUTING:
                FService.DoYield;
            OCI_NO_DATA, OCI_NEED_DATA, OCI_SUCCESS:
                Check('OCIStmtFetch', OCI_SUCCESS);
            else
                Check('OCIStmtFetch', res);
        end;
        DoPiecedFetch(pieceLoopData, res);
    until (res <> OCI_NEED_DATA) and (res <> OCI_STILL_EXECUTING);
    FLastRowCount := ROW_COUNT - prevRowCount;
    FEOF := FLastRowCount <> ANRows;
end;

procedure TOCIStatement.Next(ANRows: sb4);
begin
    Fetch(ANRows, OCI_FETCH_NEXT);
end;

procedure TOCIStatement.CancelCursor;
begin
    Check('OCIStmtFetch', NCOci.OCIStmtFetch(FHandle, Error.FHandle, 0,
        OCI_FETCH_NEXT, OCI_DEFAULT));
end;

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------

const
    nc2ociValueType: array[TOCIVarDataType] of ub2 = (0, SQLT_INT, SQLT_INT,
        SQLT_FLT, SQLT_FLT, SQLT_CHR, SQLT_AFC, SQLT_BIN, SQLT_DAT, SQLT_LNG,
        SQLT_LBI, SQLT_CUR, SQLT_CLOB, SQLT_BLOB, SQLT_CFILEE, SQLT_BFILEE,
        SQLT_CHR);
    nc2ociValueSize: array[TOCIVarDataType] of sb4 = (0, SizeOf(Integer),
        SizeOf(Integer), SizeOf(Double), SizeOf(Double), OCI_DEF_INITSTRSIZE,
        OCI_DEF_INITSTRSIZE, OCI_DEF_INITSTRSIZE, SizeOf(TOraDate),
        OCI_DEF_INITLONGSIZE, OCI_DEF_INITLONGSIZE, sizeof(pOCIStmt),
        sizeof(pOCILobLocator), sizeof(pOCILobLocator), sizeof(pOCILobLocator),
        sizeof(pOCILobLocator), OCI_DEF_ROWIDLEN);
type
    PDateTimeRec = ^TDateTimeRec;
    TDateTimeRec = record
        DateTime: TDateTime;
    end;

procedure OraDate2MSecs(pIn: pUb1; var DR: TDateTimeRec);
var
    D: TDateTime;
begin
    D := 0.0;
    with POraDate(pIn)^ do
        if Century <> 0 then
            D := EncodeDate((Century - 100) * 100 + (Year - 100), Month, Day) +
                 EncodeTime(Hour - 1, Minute - 1, Second - 1, 0);
    DR.DateTime := TimeStampToMSecs(DateTimeToTimeStamp(D));
end;

procedure MSecs2OraDate(pOut: pUb1; DR: TDateTimeRec);
var
    tmp, Y, M, D2, H, M2, S: Word;
    D: TDateTime;
begin
    D := TimeStampToDateTime(MSecsToTimeStamp(DR.DateTime));
    with POraDate(pOut)^ do begin
        DecodeDate(D, Y, M, D2);
        Century := Y div 100 + 100;
        Year := Y - (Century - 100) * 100 + 100;
        Month := M;
        Day := D2;
        DecodeTime(D, H, M2, S, tmp);
        Hour := H + 1;
        Minute := M2 + 1;
        Second := S + 1;
    end;
end;

// -------------------------------------------------
// create & destroy & prop's

constructor TOCIVariable.Create(AStmt: TOCIStatement);
begin
    inherited Create;
    FOwner := AStmt;
    ArrayLen := 1;
    InrecDataSize := OCI_DEF_INRECDATASIZE;
end;

destructor TOCIVariable.Destroy;
begin
    FreeBuffer;
    FHandle := nil;
    inherited Destroy;
end;

procedure TOCIVariable.SetArrayLen(const Value: sb4);
begin
    if FMaxArr_len <> Value then begin
        FreeBuffer;
        FMaxArr_len := Value;
        FCurEle := Value;
    end;
end;

procedure TOCIVariable.SetDataSize(const Value: sb4);
begin
    if FValue_sz <> Value then begin
        FreeBuffer;
        FValue_sz := Value;
        InitLongData;
    end;
end;

procedure TOCIVariable.SetDataType(const Value: TOCIVarDataType);
begin
    if FDataType <> Value then begin
        FreeBuffer;
        FDataType := Value;
        FValue_ty := nc2ociValueType[Value];
        FValue_sz := nc2ociValueSize[Value];
        InitLongData;
    end;
end;

procedure TOCIVariable.SetVarType(const Value: TOCIVarType);
begin
    if FVarType <> Value then begin
        FreeBuffer;
        FVarType := Value;
        InitLongData;
    end;
end;

// -------------------------------------------------
// Bind record's array access

function TOCIVariable.GetBindRecLen: sb4;
begin
    Result := SizeOf(TBindData);
    if LongData then
        Inc(Result, SizeOf(TOraLong))
    else
        Inc(Result, FValue_sz);
end;

function TOCIVariable.GetPBindData(AIndex: sb4): PBindData;
begin
    CheckBufferAccess(AIndex);
    Result := PBindData(Integer(FValue_p) + AIndex * GetBindRecLen);
end;

function TOCIVariable.GetBuffer(AIndex: sb4): pUb1;
begin
    Result := @(GetPBindData(AIndex)^.FBuffer);
end;

procedure TOCIVariable.ClearBuffer(AIndex: ub4 = MAXINT; ANoHandleFree: Boolean = False);
var
    i: Integer;
    dType: ub4;

    procedure DoClear(i: Integer);
    begin
        if LongData then
            FreeLong(POraLong(Buffer[i]))
        else if (dType <> 0) and not ANoHandleFree and (ppOCIHandle(Buffer[i])^ <> nil) then begin
            CheckOwner;
            Check('OCIDescriptorFree(' + hndlNames[dType] + ')',
                NCOci.OCIDescriptorFree(ppOCIHandle(Buffer[i])^, dType));
            ppOCIHandle(Buffer[i])^ := nil;
        end;
    end;

begin
    dType := 0;
    case DataType of
    otCursor:
        dType := OCI_DTYPE_RSET;
    otCLOB, otBLOB:
        dType := OCI_DTYPE_LOB;
    otCFile, otBFile:
        dType := OCI_DTYPE_FILE;
    end;
    if AIndex = MAXINT then
        for i := 0 to FMaxArr_len - 1 do
            DoClear(i)
    else
        DoClear(AIndex);
    if FBuffStatus = bsExtInit then
        FBuffStatus := bsExtUnknown;
    if FOwner <> nil then
        TOCIStatement(FOwner).RemovePieceVar(Self);
end;

procedure TOCIVariable.FreeBuffer;
begin
    if (FValue_p <> nil) and (FBuffStatus in [bsInt, bsExtInit]) then begin
        ClearBuffer(MAXINT, False);
        if FBuffStatus = bsInt then
            FreeMem(FValue_p);
        FValue_p := nil;
        FBuffStatus := bsInt;
    end;
end;

procedure TOCIVariable.InitBuffer(AIndex: ub4 = MAXINT; ANoHandleFree: Boolean = False);
var
    i: Integer;
    dType: ub4;

    procedure DoInit(i: Integer);
    begin
        with GetPBindData(i)^ do begin
            FInd := -1;
            FALen := FValue_sz;
            FRCode := OCI_SUCCESS;
        end;
        if LongData then
            InitLong(POraLong(Buffer[i]))
        else if (dType <> 0) and not ANoHandleFree then begin
            ppOCIHandle(Buffer[i])^ := nil;
            CheckOwner;
            Check('OCIDescriptorAlloc(' + hndlNames[dType] + ')',
                NCOci.OCIDescriptorAlloc(FOwner.FOwner.FHandle,
                ppOCIHandle(Buffer[i])^, dType, 0, nil));
        end;
    end;

begin
    dType := 0;
    case DataType of
    otCursor:
        dType := OCI_DTYPE_RSET;
    otCLOB, otBLOB:
        dType := OCI_DTYPE_LOB;
    otCFile, otBFile:
        dType := OCI_DTYPE_FILE;
    end;
    if AIndex = MAXINT then
        for i := 0 to FMaxArr_len - 1 do
            DoInit(i)
    else
        DoInit(AIndex);
    if FBuffStatus = bsExtUnknown then
        FBuffStatus := bsExtInit;
end;

procedure TOCIVariable.CheckOwner;
begin
    if (FOwner = nil) then
        OCIDBError(msgOCIVarWithoutStmt);
end;

procedure TOCIVariable.CheckBufferAccess(AIndex: sb4);
begin
    if AIndex >= FMaxArr_len then
        OCIDBError(msgOCIOutOfCount);
    if (FBuffStatus <> bsExtInit) and (
           (FValue_p = nil) and (FBuffStatus = bsInt) or
           (FValue_p <> nil) and (FBuffStatus = bsExtUnknown)
       ) then begin
        if FValue_ty <= 0 then
            OCIDBErrorFmt(msgOCIBadValueType, [FPlaceHolder, FValue_ty]);
        if FValue_sz <= 0 then
            OCIDBError(msgOCIBadValueSize);
        if FMaxArr_len <= 0 then
            OCIDBError(msgOCIBadArrayLen);
        if FBuffStatus = bsInt then
            GetMem(FValue_p, GetBindRecLen * FMaxArr_len);
        InitBuffer(MAXINT, False);
    end;
end;

procedure TOCIVariable.CheckCanWrite;
begin
    if VarType = odRet then
        OCIDBError(msgOCIDataMBAtRt);
end;

procedure TOCIVariable.SetExternBuffer(AValue: pUb1);
begin
    FValue_p := PBindData(AValue);
    FBuffStatus := bsExtInit;
end;

function TOCIVariable.GetExternBuffer: pUb1;
begin
    Result := pUb1(FValue_p);
end;

// -------------------------------------------------
// actual bind

function InCBFunc(ictxp: Pointer; bindp: pOCIBind; iter: ub4; index: ub4;
    var bufpp: Pointer; var alenp: Ub4; var piecep: Ub1; var indp: pSb2): sword; cdecl;
begin
    bufpp := nil;
    alenp := 0;
    indp := nil;
    piecep := OCI_ONE_PIECE;
    Result := OCI_CONTINUE;
end;

function OutCBFunc(ictxp: Pointer; bindp: pOCIBind; iter: ub4; index: ub4;
    var bufpp: Pointer; var alenpp: pUb4; var piecep: ub1; var indpp: pSb2;
    var rcodepp: pUb2): sword; cdecl;
var
    V: TOCIVariable;
    pBD: PBindData;
begin
    V := TOCIVariable(ictxp);
    if index = 0 then begin
        // check number of rows returned is valid
        // support ONLY for 1 row per 1 iteration
        if V.ROWS_RETURNED <> 1 then begin
            Result := OCI_ERROR;
            Exit;
        end;
    end;
    pBD := V.GetPBindData(iter);
    bufpp := @(pBD^.FBuffer);
    alenpp := @(pBD^.FALen);
    indpp := @(pBD^.FInd);
    rcodepp := @(pBD^.FRCode);
    piecep := OCI_ONE_PIECE;
    Result := OCI_CONTINUE;
end;

procedure TOCIVariable.Bind;
var
    n: ub4;
    v1: ub4;
    v2: pUb4;
begin
    CheckOwner;
    CheckBufferAccess(0);
    if VarType = odDefine then begin
        FType := OCI_HTYPE_DEFINE;
        if LongData then begin
            Check('OCIDefineByPos(' + IntToStr(FPosition) + ')',
                NCOci.OCIDefineByPos(FOwner.FHandle, FHandle, Error.FHandle,
                FPosition, nil, FValue_sz, FValue_ty, nil, nil, nil, OCI_DYNAMIC_FETCH));
            TOCIStatement(FOwner).AddPieceVar(Self);
        end
        else begin
            Check('OCIDefineByPos(' + IntToStr(FPosition) + ')',
                NCOci.OCIDefineByPos(FOwner.FHandle, FHandle, Error.FHandle,
                FPosition, @(FValue_p^.FBuffer), FValue_sz, FValue_ty, @(FValue_p^.FInd),
                @(FValue_p^.FALen), @(FValue_p^.FRCode), OCI_DEFAULT));
            n := GetBindRecLen;
            Check('OCIDefineArrayOfStruct', NCOci.OCIDefineArrayOfStruct(FHandle,
                Error.FHandle, n, n, n, n));
        end;
    end
    else if VarType = odRet then begin
        FType := OCI_HTYPE_BIND;
        Check('OCIBindByName(' + FPlaceHolder + ')', NCOci.OCIBindByName(FOwner.FHandle,
            FHandle, Error.FHandle, PChar(FPlaceHolder), Length(FPlaceHolder), nil,
            FValue_sz, FValue_ty, nil, nil, nil, 0, nil, OCI_DATA_AT_EXEC));
        Check('OCIBindDynamic', NCOci.OCIBindDynamic(FHandle, Error.FHandle,
            self, @InCBFunc, self, @OutCBFunc));
    end
    else if VarType in [odIn, odOut, odInOut] then begin
        FType := OCI_HTYPE_BIND;
        if FIsPLSQLTable and
           (TOCIStatement(FOwner).Stmt_Type in [OCI_STMT_BEGIN, OCI_STMT_DECLARE]) then begin
            v1 := FMaxArr_len;
            v2 := @FCurEle;
        end
        else begin
            v1 := 0;
            v2 := nil;
        end;
        if FPosition <> 0 then
            Check('OCIBindByPos(' + IntToStr(FPosition) + ')',
                NCOci.OCIBindByPos(FOwner.FHandle, FHandle, Error.FHandle, FPosition,
                @(FValue_p^.FBuffer), FValue_sz, FValue_ty, @(FValue_p^.FInd),
                @(FValue_p^.FALen), @(FValue_p^.FRCode), v1, v2, OCI_DEFAULT))
        else
            Check('OCIBindByName(' + FPlaceHolder + ')', NCOci.OCIBindByName(FOwner.FHandle,
                FHandle, Error.FHandle, PChar(FPlaceHolder), Length(FPlaceHolder),
                @(FValue_p^.FBuffer), FValue_sz, FValue_ty, @(FValue_p^.FInd),
                @(FValue_p^.FALen), @(FValue_p^.FRCode), v1, v2, OCI_DEFAULT));
        n := GetBindRecLen;
        Check('OCIBindArrayOfStruct', NCOci.OCIBindArrayOfStruct(FHandle,
            Error.FHandle, n, n, n, n));
    end
    else
        OCIDBError(msgOCIBadVarType);
end;

procedure TOCIVariable.BindTo(AStmt: TOCIStatement);
begin
    FOwner := AStmt;
    Bind;
end;

procedure TOCIVariable.BindOff;
begin
    if FOwner <> nil then begin
        ClearBuffer;
        TOCIStatement(FOwner).RemovePieceVar(Self);
        FOwner := nil;
        FHandle := nil;
        FError := nil;
    end;
end;

// -------------------------------------------------
// long & long raw value

procedure TOCIVariable.SetInrecDataSize(const Value: sb4);
begin
    if FInrecDataSize <> Value then begin
        FreeBuffer;
        FInrecDataSize := Value;
        InitLongData;
    end;
end;

procedure TOCIVariable.InitLongData;
begin
    FLongData := (FDataType in [otString, otChar, otRaw, otLong, otLongRaw]) and
                 (FVarType in [odDefine, odOut, odInOut]) and
                 (FValue_sz > FInrecDataSize);
end;

procedure TOCIVariable.FreeLong(pLong: POraLong);
begin
    if pLong^.FData <> nil then
        try
            FreeMem(pLong^.FData, pLong^.FSize);
        finally
            InitLong(pLong);
        end;
end;

procedure TOCIVariable.InitLong(pLong: POraLong);
begin
    pLong^.FData := nil;
    pLong^.FSize := 0;
end;

procedure TOCIVariable.GetLong(pLong: POraLong; var AData: pUb1; var ASize: sb4);
begin
    ASize := pLong^.FSize;
    AData := pLong^.FData;
end;

procedure TOCIVariable.SetLong(pLong: POraLong; AData: pUb1; ASize: sb4);
begin
    FreeLong(pLong);
    GetMem(pLong^.FData, ASize);
    Move(AData^, pLong^.FData^, ASize);
    pLong^.FSize := ASize;
end;

procedure TOCIVariable.AppendLong(pLong: POraLong; AData: pUb1; ASize: sb4);
begin
    if pLong^.FData = nil then
        SetLong(pLong, AData, ASize)
    else begin
        ReallocMem(pLong^.FData, pLong^.FSize + ASize);
        Move(AData^, pUb1(Integer(pLong^.FData) + pLong^.FSize)^, ASize);
        Inc(pLong^.FSize, ASize);
    end;
end;

// -------------------------------------------------
// get/set value

function TOCIVariable.GetDataPtr(AIndex: ub4; var ABuff: pUb1; var ASize: sb4): Boolean;
begin
    with GetPBindData(AIndex)^ do
        if FInd = -1 then begin
            Result := False;
            if DataType in [otCursor, otCLOB, otBLOB, otCFile, otBFile] then
                ABuff := @FBuffer
            else
                ABuff := nil;
        end
        else begin
            Result := True;
            ABuff := @FBuffer;
            case DataType of
                otSmallInt, otInteger:
                    ASize := SizeOf(Integer);
                otFloat, otBCD:
                    ASize := SizeOf(Double);
                otString, otChar, otRaw, otLong, otLongRaw, otROWID:
                    if LongData then
                        GetLong(POraLong(ABuff), ABuff, ASize)
                    else
                        ASize := FALen;
                otDateTime:
                    ASize := 7;
                otCursor, otCLOB, otBLOB, otCFile, otBFile:
                    ASize := SizeOf(pOCIHandle);
            else
                OCIDBError(msgOCIUnsupValueType);
            end;
        end
end;

function TOCIVariable.GetData(AIndex: ub4; ABuff: pUb1; var ASize: sb4): Boolean;
var
    p: pUb1;
begin
    Result := GetDataPtr(AIndex, p, ASize);
    if (ABuff <> nil) and
       (Result or (DataType in [otCursor, otCLOB, otBLOB, otCFile, otBFile])) then
        case DataType of
            otSmallInt:
                PSmallInt(ABuff)^ := PInteger(p)^;
            otInteger:
                PInteger(ABuff)^ := PInteger(p)^;
            otFloat:
                PDouble(ABuff)^ := PDouble(p)^;
            otBCD:
                PCurrency(ABuff)^ := PDouble(p)^;
            otString, otChar, otRaw, otLong, otLongRaw, otROWID:
                Move(p^, PChar(ABuff)^, ASize);
            otDateTime:
                begin
                    OraDate2MSecs(p, PDateTimeRec(ABuff)^);
                    ASize := SizeOf(TDateTime);
                end;
            otCursor, otCLOB, otBLOB, otCFile, otBFile:
                ppOCIHandle(ABuff)^ := ppOCIHandle(p)^;
        end;
end;

procedure TOCIVariable.SetIsNull(AIndex: ub4; AIsNull: Boolean);
begin
    with GetPBindData(AIndex)^ do
        if AIsNull then
            FInd := -1
        else
            FInd := 0;
end;

procedure TOCIVariable.AppendData(AIndex: ub4; ABuff: pUb1; ASize: sb4);
begin
    CheckCanWrite;
    if not LongData then
        OCIDBError(msgOCIUnsupValueType);
    with GetPBindData(AIndex)^ do begin
        AppendLong(POraLong(@FBuffer), ABuff, ASize);
        FInd := 0;
        FALen := SizeOf(TOraLong);
    end;
end;

procedure TOCIVariable.SetData(AIndex: ub4; ABuff: pUb1; ASize: sb4);
var
    init: LongBool;
begin
    CheckCanWrite;
    with GetPBindData(AIndex)^ do begin
        if LongData then
            FreeLong(POraLong(@FBuffer));
        if ABuff = nil then
            FInd := -1
        else begin
            FInd := 0;
            case DataType of
                otSmallInt:
                    begin
                        PInteger(@FBuffer)^ := PSmallInt(ABuff)^;
                        FALen := SizeOf(Integer);
                    end;
                otInteger:
                    begin
                        PInteger(@FBuffer)^ := PInteger(ABuff)^;
                        FALen := SizeOf(Integer);
                    end;
                otFloat:
                    begin
                        PDouble(@FBuffer)^ := PDouble(ABuff)^;
                        FALen := Sizeof(Double);
                    end;
                otBCD:
                    begin
                        PDouble(@FBuffer)^ := PCurrency(ABuff)^;
                        FALen := Sizeof(Double);
                    end;
                otString, otChar, otRaw, otLong, otLongRaw, otROWID:
                    begin
                        if ASize = -1 then
                            ASize := StrLen(PChar(ABuff));
                        if ASize > FValue_sz then
                            OCIDBError(msgOCIDataToLarge);
                        if LongData then begin
                            SetLong(POraLong(@FBuffer), ABuff, ASize);
                            FALen := SizeOf(TOraLong);
                        end
                        else begin
                            Move(PChar(ABuff)^, FBuffer, ASize);
                            FALen := ASize;
                        end;
                    end;
                otDateTime:
                    begin
                        if (ASize = 0) or (ASize = SizeOf(TDateTime)) then
                            MSecs2OraDate(@FBuffer, PDateTimeRec(ABuff)^)
                        else if ASize = SizeOf(TOraDate) then
                            POraDate(@FBuffer)^ := POraDate(ABuff)^
                        else
                            OCIDBError(msgOCIUnsupValueType);
                        FALen := 7;
                    end;
                otCursor, otCLOB, otBLOB, otCFile, otBFile:
                    begin
                        ppOCIHandle(@FBuffer)^ := ppOCIHandle(ABuff)^;
                        FALen := SizeOf(pOCIHandle);
                        if DataType in [otCLOB, otBLOB, otCFile, otBFile] then begin
                            Check('OCILobLocatorIsInit', NCOci.OCILobLocatorIsInit(FOwner.FOwner.FHandle,
                                Error.FHandle, ppOCIHandle(@FBuffer)^, init));
                            if not init then
                                FInd := -1;
                        end;
                    end;
            else
                OCIDBError(msgOCIUnsupValueType);
            end;
        end;
    end;
end;

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------

function TOCIDescriptor.Init(AEnv: TOCIHandle; AType: ub4): pOCIDescriptor;
begin
    Check('OCIDescriptorAlloc(' + hndlNames[AType] + ')',
        NCOci.OCIDescriptorAlloc(AEnv.FHandle, FHandle, AType, 0, nil));
    FType := AType;
    Result := @FHandle;
end;

destructor TOCIDescriptor.Destroy;
begin
    if FHandle <> nil then
        Check('OCIDescriptorFree(' + hndlNames[FType] + ')',
            NCOci.OCIDescriptorFree(FHandle, FType));
end;

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------

constructor TOCISelectItem.Create(AStmt: TOCIStatement);
begin
    inherited Create;
    FOwner := AStmt;
    FType := OCI_DTYPE_PARAM;
    FEnableInteger := False;
    FEnableBCD := False;
    FEnableLongString := False;
    FEnableFixedString := False;
end;

destructor TOCISelectItem.Destroy;
begin
    FHandle := nil;
    inherited Destroy;
end;

procedure TOCISelectItem.Bind;
begin
    Check('OCIParamGet(' + IntToStr(FPosition) + ')',
        NCOci.OCIParamGet(FOwner.FHandle, OCI_HTYPE_STMT,
        Error.FHandle, FHandle, FPosition));
end;

function TOCISelectItem.GetDataType: TOCIVarDataType;
begin
    case DATA_TYPE of
    SQLT_CHR, SQLT_VCS:
        if (FEnableLongString or (DATA_SIZE <= OCI_DEF_MAXDLP2STRSIZE)) and
           (DATA_SIZE <= OCI_DEF_MAXDLP4STRSIZE) then
            Result := otString
        else
            Result := otLong;
    SQLT_NUM:
        if FEnableInteger and (SCALE = 0) and (PRECISION <= 5) then
            Result := otSmallInt
        else if FEnableInteger and (SCALE = 0) and (PRECISION <= 10) then
            Result := otInteger
{$IFNDEF OCI_D5}
        else if FEnableBCD then
            Result := otBCD
{$ENDIF}
        else
            Result := otFloat;
    SQLT_INT, _SQLT_PLI: Result := otInteger;
    SQLT_LNG: Result := otLong;
    SQLT_RID, SQLT_RDD: Result := otROWID;
    SQLT_DAT: Result := otDateTime;
    SQLT_BIN: Result := otRaw;
    SQLT_LBI: Result := otLongRaw;
    SQLT_AFC:
        if FEnableFixedString then
            Result := otChar
        else if FEnableLongString or (DATA_SIZE <= OCI_DEF_MAXDLP2STRSIZE) then
            Result := otString
        else
            Result := otLong;
    SQLT_CLOB: Result := otCLOB;
    SQLT_BLOB: Result := otBLOB;
    SQLT_BFILEE: Result := otBFile;
    SQLT_CFILEE: Result := otCFile;
    SQLT_CUR: Result := otCursor;
    else Result := otUnknown;
    end;
end;

function TOCISelectItem.GetDataSize: ub4;
begin
    if DataType in [otString, otChar, otRaw, otLong, otLongRaw] then
        Result := DATA_SIZE
    else
        Result := nc2ociValueSize[DataType];
end;

function TOCISelectItem.GetVarType: TOCIVarType;
begin
    Result := odUnknown;
    case IOMODE of
    OCI_TYPEPARAM_OUT: Result := odOut;
    OCI_TYPEPARAM_IN: Result := odIn;
    OCI_TYPEPARAM_INOUT: Result := odInOut;
    end;
end;

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------

constructor TOCILobLocator.Create(AService: TOCIService; AType: ub4);
begin
    inherited Create;
    FOwner := AService;
    Init(AService.FOwner, AType);
    FOwned := True;
end;

constructor TOCILobLocator.CreateUseHandle(AService: TOCIService; AHandle: pOCIHandle);
begin
    FOwner := AService;
    FHandle := AHandle;
    FOwned := False;
end;

destructor TOCILobLocator.Destroy;
begin
    if not FOwned then
        FHandle := nil;
    inherited Destroy;
end;

procedure TOCILobLocator.Assign(AFrom: TOCILobLocator);
begin
    Check('OCILobAssign', NCOci.OCILobAssign(FOwner.FOwner.FHandle, Error.FHandle,
        AFrom.FHandle, FHandle));
end;

procedure TOCILobLocator.Close;
begin
    if FOCIVersion >= cvOracle8i then
        Check('OCILobClose', NCOci.OCILobClose(FOwner.FHandle, Error.FHandle, FHandle))
    else if FType = OCI_DTYPE_FILE then
        Check('OCILobFileClose', NCOci.OCILobClose(FOwner.FHandle, Error.FHandle, FHandle));
end;

function TOCILobLocator.GetLength: sb4;
begin
    Check('OCILobGetLength', NCOci.OCILobGetLength(FOwner.FHandle, Error.FHandle,
        FHandle, ub4(Result)));
end;

function TOCILobLocator.GetIsOpen: LongBool;
begin
    if FOCIVersion >= cvOracle8i then
        Check('OCILobIsOpen', NCOci.OCILobIsOpen(FOwner.FHandle, Error.FHandle,
            FHandle, Result))
    else if FType = OCI_DTYPE_FILE then
        Check('OCILobFileIsOpen', NCOci.OCILobIsOpen(FOwner.FHandle, Error.FHandle,
            FHandle, Result))
    else
        Result := True;
end;

function TOCILobLocator.GetIsInit: LongBool;
begin
    Check('OCILobLocatorIsInit', NCOci.OCILobLocatorIsInit(FOwner.FOwner.FHandle,
        Error.FHandle, FHandle, Result));
end;

procedure TOCILobLocator.Read(ABuff: pUb1; ABuffLen: ub4; var amount: ub4; offset: ub4);
begin
    Check('OCILobRead', NCOci.OCILobRead(FOwner.FHandle, Error.FHandle,
        FHandle, amount, offset, ABuff, ABuffLen, nil, nil, 0, SQLCS_IMPLICIT));
end;

procedure TOCILobLocator.Open(AReadOnly: Boolean);
const
    bool2mode: array[Boolean] of ub1 = (OCI_LOB_READWRITE, OCI_LOB_READONLY);
begin
    if FOCIVersion >= cvOracle8i then
        Check('OCILobOpen', NCOci.OCILobOpen(FOwner.FHandle, Error.FHandle,
            FHandle, bool2mode[AReadOnly]))
    else if FType = OCI_DTYPE_FILE then
        Check('OCILobFileOpen', NCOci.OCILobOpen(FOwner.FHandle, Error.FHandle,
            FHandle, OCI_LOB_READONLY));
end;

constructor TOCIIntLocator.CreateUseHandle(AService: TOCIService; AHandle: pOCIHandle);
begin
    inherited CreateUseHandle(AService, AHandle);
    FType := OCI_DTYPE_LOB;
end;

procedure TOCIIntLocator.Append(AFrom: TOCIIntLocator);
begin
    Check('OCILobAppend', NCOci.OCILobAppend(FOwner.FHandle, Error.FHandle,
        FHandle, AFrom.FHandle));
end;

procedure TOCIIntLocator.Copy(AFrom: TOCIIntLocator; amount, dst_offset, src_offset: ub4);
begin
    Check('OCILobCopy', NCOci.OCILobCopy(FOwner.FHandle, Error.FHandle, FHandle,
        AFrom.FHandle, amount, dst_offset, src_offset));
end;

procedure TOCIIntLocator.SetBuffering(const Value: Boolean);
begin
    if FBuffering <> Value then begin
        FBuffering := Value;
        if FBuffering then
            Check('OCILobEnableBuffering', NCOci.OCILobEnableBuffering(FOwner.FHandle,
                Error.FHandle, FHandle))
        else
            Check('OCILobDisableBuffering', NCOci.OCILobDisableBuffering(FOwner.FHandle,
                Error.FHandle, FHandle));
    end;
end;

procedure TOCIIntLocator.Erase(var amount: ub4; offset: ub4);
begin
    Check('OCILobErase', NCOci.OCILobErase(FOwner.FHandle, Error.FHandle,
        FHandle, amount, offset));
end;

procedure TOCIIntLocator.FlushBuffer;
begin
    Check('OCILobFlushBuffer', NCOci.OCILobFlushBuffer(FOwner.FHandle, Error.FHandle,
        FHandle, OCI_LOB_BUFFER_NOFREE));
end;

procedure TOCIIntLocator.LoadFromFile(AFrom: TOCIExtLocator; amount, dst_offset, src_offset: ub4);
begin
    Check('OCILobLoadFromFile', NCOci.OCILobLoadFromFile(FOwner.FHandle, Error.FHandle,
        FHandle, AFrom.FHandle, amount, dst_offset, src_offset));
end;

procedure TOCIIntLocator.Trim(ANewLen: ub4);
begin
    Check('OCILobTrim', NCOci.OCILobTrim(FOwner.FHandle, Error.FHandle,
        FHandle, ANewLen));
end;

procedure TOCIIntLocator.Write(ABuff: pUb1; ABuffLen: ub4; var amount: ub4; offset: ub4);
begin
    Check('OCILobWrite', NCOci.OCILobWrite(FOwner.FHandle, Error.FHandle,
        FHandle, amount, offset, ABuff, ABuffLen, OCI_ONE_PIECE, nil,
        nil, 0, SQLCS_IMPLICIT));
end;

constructor TOCIExtLocator.CreateUseHandle(AService: TOCIService; AHandle: pOCIHandle);
begin
    inherited CreateUseHandle(AService, AHandle);
    FType := OCI_DTYPE_FILE;
end;

function TOCIExtLocator.GetFileExists: Boolean;
begin
    Check('OCILobFileExists', NCOci.OCILobFileExists(FOwner.FHandle, Error.FHandle,
        FHandle, Result));
end;

procedure TOCIExtLocator.GetFileDir(var ADir, AFileName: String);
var
    dirBuff: array[0..29] of char;
    fileBuff: array[0..255] of char;
    dirLen, fileLen: ub2;
begin
    dirLen := 30;
    fileLen := 255;
    Check('OCILobFileGetName', NCOci.OCILobFileGetName(FOwner.FOwner.FHandle,
        Error.FHandle, FHandle, @dirBuff, dirLen, @fileBuff, fileLen));
    SetString(ADir, dirBuff, dirLen);
    SetString(AFileName, fileBuff, fileLen);
end;

procedure TOCIExtLocator.SetFileDir(const ADir, AFileName: String);
begin
    Check('OCILobFileSetName', NCOci.OCILobFileSetName(FOwner.FOwner.FHandle,
        Error.FHandle, FHandle, PChar(ADir), System.Length(ADir), PChar(AFileName),
        System.Length(AFileName)));
end;

function TOCIExtLocator.GetDirectory: String;
var
    AFileName: String;
begin
    GetFileDir(Result, AFileName);
end;

function TOCIExtLocator.GetFileName: String;
var
    ADir: String;
begin
    GetFileDir(ADir, Result);
end;

procedure TOCIExtLocator.SetDirectory(const Value: String);
begin
    SetFileDir(Value, FileName);
end;

procedure TOCIExtLocator.SetFileName(const Value: String);
begin
    SetFileDir(Directory, Value);
end;

// --------------------------------------------------------------------------
// --------------------------------------------------------------------------

constructor TOCIDescribe.Create(ASvc: TOCIService);
begin
    inherited Create;
    FOwner := ASvc;
    Init(ASvc.FOwner, OCI_HTYPE_DESCRIBE);
    FStack := TList.Create;
    FCurr := TOCIHandle.Create;
    FCurr.FOwner := FOwner;
    FCurr.FType := OCI_DTYPE_PARAM;
end;

destructor TOCIDescribe.Destroy;
begin
    FStack.Free;
    FCurr.FHandle := nil;
    FCurr.Free;
    inherited Destroy;
end;

procedure TOCIDescribe.DescribeName(const AName: String);
begin
    // ??? work in nonblocking mode
    Check('OCIDescribeAny', NCOci.OCIDescribeAny(FOwner.FHandle, Error.FHandle,
        PChar(AName), Length(AName), OCI_OTYPE_NAME, 0, OCI_PTYPE_UNK, FHandle));
    FCurr.FHandle := GetHandleAttr(OCI_ATTR_PARAM);
end;

function TOCIDescribe.OpenList(AAtrType: Integer): Integer;
begin
    FStack.Add(FCurr.FHandle);
    FStack.Add(FCurr.GetHandleAttr(AAtrType));
    Check('OCIAttrGet ub2', NCOci.OCIAttrGet(FStack.Last, OCI_DTYPE_PARAM,
        @Result, nil, OCI_ATTR_NUM_PARAMS, Error.FHandle));
end;

procedure TOCIDescribe.CloseList;
begin
    FCurr.FHandle := FStack.Items[FStack.Count - 2];
    FStack.Delete(FStack.Count - 1);
    FStack.Delete(FStack.Count - 1);
end;

procedure TOCIDescribe.GoToItem(AIndex: Integer);
begin
    Check('OCIParamGet', NCOci.OCIParamGet(FStack.Last, OCI_DTYPE_PARAM,
        Error.FHandle, FCurr.FHandle, AIndex));
end;

function TOCIDescribe.GetSelectItem(AIndex: Integer): TOCISelectItem;
begin
    GoToItem(AIndex);
    Result := TOCISelectItem.Create(TOCIStatement(FOwner));
    Result.FHandle := FCurr.FHandle;
end;

function TOCIDescribe.GetPtr(AIndex: Integer): Pointer;
begin
    Result := FCurr.GetHandleAttr(AIndex);
end;

function TOCIDescribe.GetSB1(AIndex: Integer): sb1;
begin
    Result := FCurr.GetSB1Attr(AIndex);
end;

function TOCIDescribe.GetText(AIndex: Integer): String;
begin
    Result := FCurr.GetStringAttr(AIndex);
end;

function TOCIDescribe.GetUB1(AIndex: Integer): ub1;
begin
    Result := FCurr.GetUB1Attr(AIndex);
end;

function TOCIDescribe.GetUB2(AIndex: Integer): ub2;
begin
    Result := FCurr.GetUB2Attr(AIndex);
end;

function TOCIDescribe.GetUB4(AIndex: Integer): ub4;
begin
    Result := FCurr.GetUB4Attr(AIndex);
end;

initialization

    FGlobalPieceBuffUseCnt := 0;

end.
